# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: metagriffin <mg.github@metagriffin.net>
# date: 2016/06/09
# copy: (C) Copyright 2016-EOT metagriffin -- see LICENSE.txt
#------------------------------------------------------------------------------
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#------------------------------------------------------------------------------

import bcrypt
import logging

# TODO: cache the htpasswd file... **BUT** only do that if there is
#       also a check for if the file changes

#------------------------------------------------------------------------------

log = logging.getLogger(__name__)

#------------------------------------------------------------------------------
class Htpasswd(object):

  #----------------------------------------------------------------------------
  def __init__(self, filename, *args, **kw):
    super(Htpasswd, self).__init__(*args, **kw)
    self.filename = filename

  #----------------------------------------------------------------------------
  def _getEntries(self):
    with open(self.filename, 'rb') as fp:
      data = fp.read().strip().split('\n')
      data = [item.split(':', 1)
              for item in data
              if item and not item.startswith('##')]
      return dict(data)

  #----------------------------------------------------------------------------
  def isUser(self, username):
    return username in self._getEntries()

  #----------------------------------------------------------------------------
  def check(self, username, credential):
    check = self._getEntries().get(username, None)
    if check is None:
      log.warning('check for unknown user %r', username)
      return False
    if check.startswith('$2a$') or check.startswith('$2y$'):
      return check == bcrypt.hashpw(credential, check)
    log.error('unknown htpasswd hash entry for user %r', username)
    return False


#------------------------------------------------------------------------------
# end of $Id$
# $ChangeLog$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: metagriffin <mg.github@metagriffin.net>
# date: 2016/06/09
# copy: (C) Copyright 2016-EOT metagriffin -- see LICENSE.txt
#------------------------------------------------------------------------------
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#------------------------------------------------------------------------------

import sys
import struct
import logging
import time
import re

from . import htpasswd

#------------------------------------------------------------------------------

log = logging.getLogger(__name__)

#------------------------------------------------------------------------------
class Engine(object):

  #----------------------------------------------------------------------------
  def __init__(self, options, *args, **kw):
    super(Engine, self).__init__(*args, **kw)
    self.options   = options
    self.htpasswd  = htpasswd.Htpasswd(options.htpasswd)
    self.domainsed = None
    if options.domainsed:
      # TODO: this is not "rigorous"...
      src, dst, rem = options.domainsed.split(options.domainsed[0], 3)[1:]
      # todo: use `rem`
      # todo: is there a package to help me with this?...
      self.domainsed = (re.compile(src), dst)

  #----------------------------------------------------------------------------
  def run(self):
    try:
      self._run()
    except Exception as err:
      log.exception('failed during main loop')
      raise

  #----------------------------------------------------------------------------
  def _run(self):
    log.info('beginning main processing loop')
    while True:
      request = self.read()
      log.debug(
        'request: "%s[...]" (%d bytes total)', request[0:5], len(request))
      result  = self.processRequest(request)
      log.debug('response: %r', result)
      self.write(result)

  #----------------------------------------------------------------------------
  def processRequest(self, request):
    result  = 0
    params  = request.split(':')
    cmd     = params[0]
    if cmd == 'auth':
      user, domain, cred = params[1:4]
      user = self._transformUser(user, domain)
      log.debug('authenticating %r', user)
      if not self.htpasswd.check(user, cred):
        log.warning('authentication for %r failed (bad username/password)', user)
        time.sleep(4)
        return 0
      log.info('authentication for %r succeeded', user)
      return 1
    if cmd == 'isuser':
      user, domain = params[1:3]
      user = self._transformUser(user, domain)
      log.debug('checking for existence of %r', user)
      if not self.htpasswd.isUser(user):
        return 0
      return 1
    log.error('unknown/unimplemented command: %r', request)
    return 0

  #----------------------------------------------------------------------------
  def _transformUser(self, user, domain):
    if domain and self.domainsed:
      domain = self.domainsed[0].sub(self.domainsed[1], domain)
    if domain:
      return user + '@' + domain
    return user

  #----------------------------------------------------------------------------
  def read(self):
    input_length = sys.stdin.read(2)
    (size,) = struct.unpack('>h', input_length)
    return sys.stdin.read(size)

  #----------------------------------------------------------------------------
  def write(self, result):
    token = struct.pack('>hh', 2, result)
    sys.stdout.write(token)
    sys.stdout.flush()


#------------------------------------------------------------------------------
# end of $Id$
# $ChangeLog$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: metagriffin <mg.github@metagriffin.net>
# date: 2016/06/09
# copy: (C) Copyright 2016-EOT metagriffin -- see LICENSE.txt
#------------------------------------------------------------------------------
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#------------------------------------------------------------------------------

import logging
import logging.handlers
import argparse
import time

from .i18n import _
from . import engine

#------------------------------------------------------------------------------

DEFAULT_HTPASSWD        = '/var/www/.htpasswd'
DEFAULT_LOGFILE         = '/var/log/ejabberd/auth-htpasswd.log'
DEFAULT_LOGSIZE         = 1048576
DEFAULT_LOGCOUNT        = 10

#------------------------------------------------------------------------------
def main(argv=None):

  cli = argparse.ArgumentParser(
    description = _(
      'Yet another ejabberd authentication bridge that uses apache-style'
      ' .htpasswd files.'),
  )

  cli.add_argument(
    _('-d'), _('--debug'),
    dest='debug', action='store_true', default=False,
    help=_('enable debugging output'))

  cli.add_argument(
    _('-l'), _('--log-file'), metavar=_('FILENAME'),
    dest='logfile', action='store', default=DEFAULT_LOGFILE,
    help=_('specify the logging file; default: "%(default)s"'))

  cli.add_argument(
    _('-s'), _('--log-size'), metavar=_('BYTES'),
    dest='logsize', action='store', type=int, default=DEFAULT_LOGSIZE,
    help=_('specify the maximum size of the logging file;'
           ' default: %(default)s'))

  cli.add_argument(
    _('-c'), _('--log-count'), metavar=_('NUMBER'),
    dest='logcount', action='store', type=int, default=DEFAULT_LOGCOUNT,
    help=_('specify the maximum number of logging rotation files;'
           ' default: %(default)s'))

  cli.add_argument(
    _('-t'), _('--domain-transform'), metavar=_('EXPR'),
    dest='domainsed', action='store', default=None,
    help=_('specify a "sed" substitution expression for domain name'
           ' transformation; example:'
           ' "-t /chat\\.example\\.com/example\\.com/"'))

  cli.add_argument(
    'htpasswd', metavar=_('FILENAME'),
    nargs='?', default=DEFAULT_HTPASSWD,
    help=_('path to apache-style ".htpasswd" file; default: "%(default)s"'))

  options = cli.parse_args(args=argv)

  # todo: i should really do this:
  #         logging.config.fileConfig(options.logconfig)
  #       ...

  logging.basicConfig()

  root = logging.getLogger()
  for handler in root.handlers:
    root.removeHandler(handler)
  handler = logging.handlers.RotatingFileHandler(
    options.logfile,
    maxBytes    = options.logsize,
    backupCount = options.logcount,
  )
  handler.setFormatter(logging.Formatter(
    fmt       = '[%(asctime)s] %(levelname)-5.5s [%(name)s,pid=%(process)d] %(message)s',
    datefmt   = '%Y-%m-%dT%H:%M:%SZ',
  ))
  handler.formatter.converter = time.gmtime
  root.addHandler(handler)
  root.setLevel(logging.DEBUG if options.debug else logging.INFO)

  try:
    proc = engine.Engine(options)
  except Exception:
    root.exception('failed starting ejabberd-htpasswd engine')
  proc.run()

#------------------------------------------------------------------------------
# end of $Id$
# $ChangeLog$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: metagriffin <mg.github@metagriffin.net>
# date: 2016/06/09
# copy: (C) Copyright 2013-EOT metagriffin -- see LICENSE.txt
#------------------------------------------------------------------------------
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#------------------------------------------------------------------------------

from gettext import gettext

#------------------------------------------------------------------------------
def _(message, *args, **kw):
  if args or kw:
    return gettext(message).format(*args, **kw)
  return gettext(message)

#------------------------------------------------------------------------------
# end of $Id$
# $ChangeLog$
#------------------------------------------------------------------------------
